from enum import IntEnum


class Outcome(IntEnum):
    """
    Enum of evaluation outcomes. A candidate solution can be a new global best,
    a better solution than the current incumbent, just accepted (but not
    improving anything), or rejected.
    """

    BEST = 0  #: Candidate solution is a new global best.
    BETTER = 1  #: Candidate solution is better than the current incumbent.
    ACCEPT = 2  #: Candidate solution is accepted.
    REJECT = 3  #: Candidate solution is rejected.



import numpy as np

from alns.select.OperatorSelectionScheme import OperatorSelectionScheme


class RandomSelect(OperatorSelectionScheme):
    """
    Randomly selects operator pairs with uniform probability. The operator
    pairs respect the operator coupling matrix.
    """

    def __call__(self, rnd, best, curr):
        """
        Selects a (destroy, repair) operator pair with uniform probability.
        """
        allowed = np.argwhere(self._op_coupling)
        idx = rnd.randint(len(allowed))

        return tuple(allowed[idx])

    def update(self, candidate, d_idx, r_idx, outcome):
        pass  # pragma: no cover



import logging
from typing import List, Optional

import numpy as np

from alns.select.RouletteWheel import RouletteWheel
from alns.State import State

logger = logging.getLogger(__name__)


class SegmentedRouletteWheel(RouletteWheel):
    """
    An operator selection scheme based on the roulette wheel mechanism, where
    each operator is selected based on normalised weights. Weights are not
    updated in each iteration, but only after each segment. Scores are gathered
    during each segment, as:

    ``seg_weight += score``

    At the start of each segment, ``seg_weight`` is reset to zero. At the end
    of a segment, the weights are updated as:

    ``new_weight = decay * old_weight + (1 - decay) * seg_weight``

    Parameters
    ----------
    scores
        A list of four non-negative elements, representing the weight
        updates when the candidate solution results in a new global best
        (idx 0), is better than the current solution (idx 1), the solution
        is accepted (idx 2), or rejected (idx 3).
    decay
        Decay parameter in [0, 1]. This parameter is used to weigh the
        running performance of each operator.
    seg_length
        Length of a single segment.
    num_destroy
        Number of destroy operators.
    num_repair
        Number of repair operators.
    op_coupling
        Optional boolean matrix that indicates coupling between destroy and
        repair operators. Entry (i, j) is True if destroy operator i can be
        used together with repair operator j, and False otherwise.
    """

    def __init__(
        self,
        scores: List[float],
        decay: float,
        seg_length: int,
        num_destroy: int,
        num_repair: int,
        op_coupling: Optional[np.ndarray] = None,
    ):
        super().__init__(scores, decay, num_destroy, num_repair, op_coupling)

        if seg_length < 1:
            raise ValueError("seg_length < 1 not understood.")

        self._seg_length = seg_length
        self._iter = 0

        self._reset_segment_weights()

    @property
    def seg_length(self):
        return self._seg_length

    def __call__(self, rnd_state, best: State, curr: State):
        self._iter += 1

        if self._iter % self._seg_length == 0:
            logger.debug(f"End of segment (#iters = {self._iter}).")

            self._d_weights *= self._decay
            self._d_weights += (1 - self._decay) * self._d_seg_weights

            self._r_weights *= self._decay
            self._r_weights += (1 - self._decay) * self._r_seg_weights

            self._reset_segment_weights()

        return super().__call__(rnd_state, best, curr)

    def update(self, cand, d_idx, r_idx, outcome):
        self._d_seg_weights[d_idx] += self._scores[outcome]
        self._r_seg_weights[r_idx] += self._scores[outcome]

    def _reset_segment_weights(self):
        self._d_seg_weights = np.zeros_like(self._d_weights)
        self._r_seg_weights = np.zeros_like(self._r_weights)



from abc import ABC, abstractmethod


class State(ABC):
    """
    State object, which stores a solution and whose cost can be evaluated
    through its ``objective()`` member function.

    The State class is abstract - you are encouraged to subclass it to suit
    your specific problem.
    """

    @abstractmethod
    def objective(self) -> float:
        """
        Computes the state's associated objective value.
        """
        return NotImplemented



from .AlphaUCB import AlphaUCB
from .OperatorSelectionScheme import OperatorSelectionScheme
from .RandomSelect import RandomSelect
from .RouletteWheel import RouletteWheel
from .SegmentedRouletteWheel import SegmentedRouletteWheel



from typing import List

import numpy as np
import numpy.random as rnd
from numpy.testing import assert_equal, assert_raises
from pytest import mark

from alns.Outcome import Outcome
from alns.select import AlphaUCB
from alns.tests.states import Zero


@mark.parametrize(
    "scores, alpha",
    [
        ([0, 0, 0, 0], 0),
        ([0, 1, 2, 3], 0.5),
        ([5, 3, 2, 1], 1),
    ],
)
def test_properties(scores, alpha):
    select = AlphaUCB(scores, alpha, 1, 1)

    assert_equal(select.scores, scores)
    assert_equal(select.alpha, alpha)


@mark.parametrize("alpha", [1.01, -0.01, -0.5, 1.05, 1.5])
def test_raises_invalid_alpha(alpha: float):
    with assert_raises(ValueError):
        AlphaUCB([0, 0, 0, 0], alpha, 1, 1)


@mark.parametrize("alpha", np.linspace(0, 1, num=5))
def test_does_not_raise_valid_decay(alpha: float):
    AlphaUCB([0, 0, 0, 0], alpha, 1, 1)


@mark.parametrize(
    "scores, alpha, num_destroy, num_repair",
    [
        ([5, 3, 2, -1], 0.5, 1, 1),  # negative score
        ([5, 3, 2], 0.5, 1, 1),  # len(score) < 4
        ([5, 3, 2, 1], -0.1, 1, 1),  # alpha < 0
        ([5, 3, 2, 1], 1.1, 1, 1),  # alpha > 1
    ],
)
def test_raises_invalid_arguments(
    scores: List[float],
    alpha: float,
    num_destroy: int,
    num_repair: int,
):
    with assert_raises(ValueError):
        AlphaUCB(scores, alpha, num_destroy, num_repair)


def test_call_with_only_one_operator_pair():
    # Only one operator pair, so the algorithm should select (0, 0).
    select = AlphaUCB([2, 1, 1, 0], 0.5, 1, 1)
    state = rnd.RandomState()

    selected = select(state, Zero(), Zero())
    assert_equal(selected, (0, 0))


def test_update_with_two_operator_pairs():
    select = AlphaUCB([2, 1, 1, 0], 0.5, 2, 1)
    state = rnd.RandomState()

    # Avg. reward for (0, 0) after this is 2, for (1, 0) is still 1 (default).
    select.update(Zero(), 0, 0, outcome=Outcome.BEST)

    # So now (0, 0) is selected again.
    selected = select(state, Zero(), Zero())
    assert_equal(selected, (0, 0))

    # One more update. Avg. reward goes to 1, and number of times to 2.
    select.update(Zero(), 0, 0, outcome=Outcome.REJECT)

    # The Q value of (0, 0) is now approx 1.432, and that of (1, 0) is now
    # approx 1.74. So (1, 0) is selected.
    selected = select(state, Zero(), Zero())
    assert_equal(selected, (1, 0))



from typing import List, Optional, Tuple

import numpy as np
from numpy.random import RandomState

from alns.select.OperatorSelectionScheme import OperatorSelectionScheme
from alns.State import State


class RouletteWheel(OperatorSelectionScheme):
    """
    An operator selection scheme based on the roulette wheel mechanism, where
    each operator is selected based on normalised weights. The operator weights
    are adjusted continuously throughout the algorithm runs. This works as
    follows. In each iteration, the old weight is updated with a score based
    on a convex combination of the existing weight and the new score, as:

    ``new_weight = decay * old_weight + (1 - decay) * score``

    Parameters
    ----------
    scores
        A list of four non-negative elements, representing the weight
        updates when the candidate solution results in a new global best
        (idx 0), is better than the current solution (idx 1), the solution
        is accepted (idx 2), or rejected (idx 3).
    decay
        Decay parameter in [0, 1]. This parameter is used to weigh the
        running performance of each operator.
    num_destroy
        Number of destroy operators.
    num_repair
        Number of repair operators.
    op_coupling
        Optional boolean matrix that indicates coupling between destroy and
        repair operators. Entry (i, j) is True if destroy operator i can be
        used together with repair operator j, and False otherwise.
    """

    def __init__(
        self,
        scores: List[float],
        decay: float,
        num_destroy: int,
        num_repair: int,
        op_coupling: Optional[np.ndarray] = None,
    ):
        super().__init__(num_destroy, num_repair, op_coupling)

        if any(score < 0 for score in scores):
            raise ValueError("Negative scores are not understood.")

        if len(scores) < 4:
            # More than four is OK because we only use the first four.
            raise ValueError(f"Expected four scores, found {len(scores)}")

        if not (0 <= decay <= 1):
            raise ValueError("decay outside [0, 1] not understood.")

        self._scores = scores
        self._d_weights = np.ones(num_destroy, dtype=float)
        self._r_weights = np.ones(num_repair, dtype=float)
        self._decay = decay

    @property
    def scores(self) -> List[float]:
        return self._scores

    @property
    def destroy_weights(self) -> np.ndarray:
        return self._d_weights

    @property
    def repair_weights(self) -> np.ndarray:
        return self._r_weights

    @property
    def decay(self) -> float:
        return self._decay

    def __call__(
        self, rnd_state: RandomState, best: State, curr: State
    ) -> Tuple[int, int]:
        """
        Selects a destroy and repair operator pair to apply in this iteration.
        The probability of an operator being selected is based on the operator
        weights: operators that frequently improve the current solution - and
        thus have higher weights - are selected with a higher probability.

        Parameters
        ----------
        rnd_state
            Random state object, to be used for random number generation.
        best
            The best solution state observed so far.
        current
            The current solution state.

        Returns
        -------
        A tuple of (d_idx, r_idx), which are indices into the destroy and
        repair operator lists, respectively.
        """

        def select(op_weights):
            probs = op_weights / np.sum(op_weights)
            return rnd_state.choice(range(len(op_weights)), p=probs)

        d_idx = select(self._d_weights)
        coupled_r_idcs = np.flatnonzero(self.op_coupling[d_idx])
        r_idx = coupled_r_idcs[select(self._r_weights[coupled_r_idcs])]

        return d_idx, r_idx

    def update(self, cand, d_idx, r_idx, outcome):
        self._d_weights[d_idx] *= self._decay
        self._d_weights[d_idx] += (1 - self._decay) * self._scores[outcome]

        self._r_weights[r_idx] *= self._decay
        self._r_weights[r_idx] += (1 - self._decay) * self._scores[outcome]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['alns',
 'alns.accept',
 'alns.accept.tests',
 'alns.select',
 'alns.select.tests',
 'alns.stop',
 'alns.stop.tests',
 'alns.tests']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=2.2.0', 'numpy>=1.15.2']

setup_kwargs = {
    'name': 'alns',
    'version': '5.0.0',
    'description': 'A flexible implementation of the adaptive large neighbourhood search (ALNS) algorithm.',
    'long_description': '[![PyPI version](https://badge.fury.io/py/alns.svg)](https://badge.fury.io/py/alns)\n[![ALNS](https://github.com/N-Wouda/ALNS/actions/workflows/alns.yml/badge.svg)](https://github.com/N-Wouda/ALNS/actions/workflows/alns.yml)\n[![Documentation Status](https://readthedocs.org/projects/alns/badge/?version=latest)](https://alns.readthedocs.io/en/latest/?badge=latest)\n[![codecov](https://codecov.io/gh/N-Wouda/ALNS/branch/master/graph/badge.svg)](https://codecov.io/gh/N-Wouda/ALNS)\n\nThis package offers a general, well-documented and tested\nimplementation of the adaptive large neighbourhood search (ALNS)\nmeta-heuristic, based on the description given in [Pisinger and Ropke\n(2010)][1]. It may be installed in the usual way as\n```\npip install alns\n```\n\n### Examples\nIf you wish to dive right in, the [documentation][7] contains example notebooks\nshowing how the ALNS library may be used. These include:\n\n- The travelling salesman problem (TSP), [here][2]. We solve an instance of 131\n  cities in one minute to a 2% optimality gap, using very simple destroy and\n  repair heuristics.\n- The capacitated vehicle routing problem (CVRP), [here][8]. We solve an \n  instance with 241 customers to within 3% of optimality using a combination\n  of a greedy repair operator, and a _slack-induced substring removal_ destroy\n  operator.\n- The cutting-stock problem (CSP), [here][4]. We solve an instance with\n  180 beams over 165 distinct sizes to within 1.35% of optimality in\n  only a very limited number of iterations.\n- The resource-constrained project scheduling problem (RCPSP), [here][6]. We solve \n  an instance with 90 jobs and 4 resources to within 4% of the best known solution,\n  using a number of different operators and enhancement techniques from the \n  literature.\n\nFinally, the features notebook gives an overview of various options available \nin the `alns` package (explained below). In the notebook we use these different\noptions to solve a toy 0/1-knapsack problem. The notebook is a good starting\npoint for when you want to use different schemes, acceptance or stopping criteria \nyourself. It is available [here][5].\n\n## How to use\nOur [documentation][7] provides a complete overview of the `alns` package. In \nshort: the `alns` package exposes two classes, `ALNS` and `State`. The first\nmay be used to run the ALNS algorithm, the second may be subclassed to\nstore a solution state - all it requires is to define an `objective`\nmember function, returning an objective value.\n\nThe ALNS algorithm must be supplied with an _operator selection scheme_, an\n_acceptance criterion_, and a _stopping criterion_. These are explained further\nin the documentation.\n\n## References\n- Pisinger, D., and Ropke, S. (2010). Large Neighborhood Search. In M.\n  Gendreau (Ed.), _Handbook of Metaheuristics_ (2 ed., pp. 399-420).\n  Springer.\n- Santini, A., Ropke, S. & Hvattum, L.M. (2018). A comparison of\n  acceptance criteria for the adaptive large neighbourhood search\n  metaheuristic. *Journal of Heuristics* 24 (5): 783-815.\n\n[1]: http://orbit.dtu.dk/en/publications/large-neighborhood-search(61a1b7ca-4bf7-4355-96ba-03fcdf021f8f).html\n[2]: https://alns.readthedocs.io/en/latest/examples/travelling_salesman_problem.html\n[3]: https://link.springer.com/article/10.1007%2Fs10732-018-9377-x\n[4]: https://alns.readthedocs.io/en/latest/examples/cutting_stock_problem.html\n[5]: https://alns.readthedocs.io/en/latest/examples/alns_features.html\n[6]: https://alns.readthedocs.io/en/latest/examples/resource_constrained_project_scheduling_problem.html\n[7]: https://alns.readthedocs.io/en/latest/\n[8]: https://alns.readthedocs.io/en/latest/examples/capacitated_vehicle_routing_problem.html\n',
    'author': 'Niels Wouda',
    'author_email': 'n.wouda@apium.nl',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/N-Wouda/ALNS',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import numpy as np
import numpy.random as rnd
from numpy.testing import assert_, assert_allclose, assert_approx_equal

from alns.select import RandomSelect
from alns.tests.states import Zero


def test_op_coupling():
    rnd_state = rnd.RandomState(1)

    # For i in {1..5}, each destroy operator i is coupled with repair operator
    # i. So only (i, i) pairs can be selected.
    op_coupling = np.eye(5)
    select = RandomSelect(5, 5, op_coupling)

    for _ in range(1_000):
        d_idx, r_idx = select(rnd_state, Zero(), Zero())
        assert_(d_idx == r_idx)


def test_uniform_selection():
    rnd_state = rnd.RandomState(1)
    histogram = np.zeros((2, 2))

    select = RandomSelect(2, 2)

    for _ in range(10_000):
        d_idx, r_idx = select(rnd_state, Zero(), Zero())
        histogram[d_idx, r_idx] += 1

    # There are four operator pair combinations, so each pair should have a
    # one in four chance of being selected. We allow a 0.01 margin since this
    # is based on sampling.
    histogram /= histogram.sum()
    assert_allclose(histogram, 0.25, atol=0.01)


def test_uniform_selection_op_coupling():
    rnd_state = rnd.RandomState(1)
    histogram = np.zeros((2, 2))

    op_coupling = np.eye(2)
    op_coupling[0, 1] = 1

    select = RandomSelect(2, 2, op_coupling)

    for _ in range(10_000):
        d_idx, r_idx = select(rnd_state, Zero(), Zero())
        histogram[d_idx, r_idx] += 1

    # There are three OK operator pair combinations, so each such pair should
    # have a one in three chance of being selected.
    histogram /= histogram.sum()

    # These should be sampled uniformly...
    assert_approx_equal(histogram[0, 0], 1 / 3, significant=2)
    assert_approx_equal(histogram[0, 1], 1 / 3, significant=2)
    assert_approx_equal(histogram[1, 1], 1 / 3, significant=2)

    # ...but this one's not allowed by the operator coupling matrix.
    assert_approx_equal(histogram[1, 0], 0, significant=7)


def test_single_operators():
    rnd_state = rnd.RandomState(1)
    select = RandomSelect(1, 1)

    # Only one (destroy, repair) operator pair, so should return (0, 0).
    assert_(select(rnd_state, Zero(), Zero()) == (0, 0))
